from pyparsing import ParserElement

ParserElement.enablePackrat()


class MetaGrammar(type):
    def __init__(cls, name, bases, dct):
        super(MetaGrammar, cls).__init__(name, bases, dct)
        try:
            cls._set_parse_action(cls.action)
        except AttributeError as e:
            pass
        try:
            cls.set_name(cls.name)
        except AttributeError as e:
            pass

    def __add__(cls, other):
        return cls.grammar.__add__(other)

    def __and__(cls, other):
        return cls.grammar.__and__(other)

    # def __eq__(cls, other):
    #     return cls.grammar.__eq__(other)

    def __getitem__(cls, key):
        return cls.grammar[key]

    def __mul__(cls, other):
        return cls.grammar.__mul__(other)

    # def __ne__(cls, other):
    #     return cls.grammar.__ne__(other)

    def __or__(cls, other):
        return cls.grammar.__or__(other)

    def __radd__(cls, other):
        return cls.grammar.__radd__(other)

    def __rand__(cls, other):
        return cls.grammar.__rand__(other)

    def __req__(cls, other):
        return cls.grammar.__req__(other)

    def __rmul__(cls, other):
        return cls.grammar.__rmul__(other)

    def __rne__(cls, other):
        return cls.grammar.__rne__(other)

    def __ror__(cls, other):
        return cls.grammar.__ror__(other)

    def __rsub__(cls, other):
        return cls.grammar.__rsub__(other)

    def __rxor__(cls, other):
        return cls.grammar.__rxor__(other)

    def __sub__(cls, other):
        return cls.grammar.__sub__(other)

    def __xor__(cls, other):
        return cls.grammar.__xor__(other)

    def get_adapter_grammar(cls):
        return cls.grammar.get_adapter_grammar()

    @property
    def _grammar(cls):
        """Return pyparsing grammar contained in this class travelling
        from MetaGrammar --> Grammar --> Adapter --> PyParsing
        """
        return cls.grammar._grammar

    def ignore(cls, expr):
        return cls._grammar.ignore(expr)

    def _set_parse_action(cls, action):
        try:
            cls.grammar._set_parse_action(action)
        except AttributeError:
            pass

    def parse(cls, string, explode=True):
        return cls.grammar.parse(string, explode)

    def set_name(cls, name):
        try:
            cls.grammar.set_name(name)
        except AttributeError:
            pass


class Grammar(metaclass=MetaGrammar):
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.value)})"

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.value == other.value

    def __ne__(self, other):
        return not self == other



from easymql.core import QuotedString, Regex, sci_real, real, signed_integer
from easymql.exc import DatePartOutOfRangeError
from easymql.keywords import null, true, false
from easymql.meta import Grammar
from easymql.utils import safe_cast_int


class PrimaryDataType(Grammar):
    pass


class Null(PrimaryDataType):

    grammar = null

    @classmethod
    def action(cls, tokens):
        return cls(None)

    def __eq__(self, other):
        if super().__eq__(other) is False:
            return isinstance(other, type(None)) and self.value is other
        return True


class String(PrimaryDataType):
    name = 'string'
    grammar = QuotedString(quoteChar='"', escChar='\\', multiline=True)

    @classmethod
    def action(cls, token):
        return cls(token[0])

    def __eq__(self, other):
        if super().__eq__(other) is False:
            return isinstance(other, str) and self.value == other
        return True


class Boolean(PrimaryDataType):

    grammar = true | false

    @classmethod
    def action(cls, token):
        return Boolean(token[0] == 'true')

    def __eq__(self, other):
        if super().__eq__(other) is False:
            return isinstance(other, bool) and self.value == other
        return True


class Integer(PrimaryDataType):
    name = 'integer'
    grammar = signed_integer

    @classmethod
    def action(cls, token):
        return cls(int(token[0]))

    def __eq__(self, other):
        if super().__eq__(other) is False:
            return isinstance(other, int) and self.value == other
        return True


class Decimal(PrimaryDataType):
    name = 'decimal'
    grammar = sci_real | real

    @classmethod
    def action(cls, token):
        return cls(token[0])

    def __eq__(self, other):
        if super().__eq__(other) is False:
            return isinstance(other, float) and self.value == other
        return True


class Number(PrimaryDataType):
    name = 'number'
    grammar = Decimal | Integer


class Date(PrimaryDataType):

    grammar = Regex(
        r'D"(?P<year>\d{4})(-(?P<month>\d{2})(-(?P<day>\d{2}))?)?'
        r'([T ](?P<hour>\d{2}):(?P<minute>\d{2})(:(?P<second>\d{2})(\.(?P<millisecond>\d{3}))?)?'
        r'(?P<timezone>(Z|[+-](?P<tzhour>\d{2})(:?(?P<tzminute>\d{2}))?))?)?"'
    )

    range_map = {
        'year': {'min': 0, 'max': 9999},
        'month': {'min': 1, 'max': 12},
        'day': {'min': 1, 'max': 31},
        'hour': {'min': 0, 'max': 23},
        'minute': {'min': 0, 'max': 59},
        'second': {'min': 0, 'max': 59},
        'millisecond': {'min': 0, 'max': 999},
        'tzhour': {'min': -26, 'max': 26},
        'tzminute': {'min': 0, 'max': 59},
    }

    def __init__(
        self,
        year,
        month=1,
        day=1,
        hour=0,
        minute=0,
        second=0,
        millisecond=0,
        timezone='Z',
    ):
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.millisecond = millisecond
        self.timezone = timezone
        self.tzhour = 0
        self.tzminute = 0
        if 4 < len(timezone) < 7:  # [+-]HH:MM or [+-]HHMM
            self.tzhour = int(timezone[:3])
            self.tzminute = int(timezone[-2:])
        elif len(timezone) == 3:  # [+-]HH
            self.tzhour = int(timezone)

        self.validate()

        super(Date, self).__init__(
            {
                '$dateFromParts': {
                    part: getattr(self, part)
                    for part in [
                        'year',
                        'month',
                        'day',
                        'hour',
                        'minute',
                        'second',
                        'millisecond',
                        'timezone',
                    ]
                }
            }
        )

    def validate(self):
        for part, range in self.range_map.items():
            value = getattr(self, part)
            min, max = range['min'], range['max']
            try:
                if value < min or value > max:
                    raise DatePartOutOfRangeError(
                        f'{part}({value}) is out of range [{min} - {max}]'
                    )
            except TypeError:
                pass

    def __repr__(self):
        arg_str = ', '.join(
            [
                repr(getattr(self, p))
                for p in [
                    'year',
                    'month',
                    'day',
                    'hour',
                    'minute',
                    'second',
                    'millisecond',
                    'timezone',
                ]
            ]
        )
        return f'Date({arg_str})'

    @classmethod
    def action(cls, tokens):
        kwargs = {
            p: safe_cast_int(tokens[p])
            for p in ['year', 'month', 'day', 'hour', 'minute', 'second', 'millisecond']
            if tokens[p] is not None
        }
        if tokens['timezone'] is not None:
            kwargs['timezone'] = tokens['timezone']
        return cls(**kwargs)


Primary = Null | String | Boolean | Number | Integer | Decimal | Date



class Action:
    @staticmethod
    def action(tokens):
        return tokens


class ExpressionAction(Action):
    @staticmethod
    def action(tokens):
        return {
            '$'
            + ''.join(
                [
                    part.capitalize() if i else part
                    for i, part in enumerate(tokens[0].lower().split('_'))
                ]
            ): tokens[1:]
        }


class UnaryExpressionAction(Action):
    @staticmethod
    def action(tokens):
        return {
            '$'
            + ''.join(
                [
                    part.capitalize() if i else part
                    for i, part in enumerate(tokens[0].lower().split('_'))
                ]
            ): tokens[-1]
        }



from easymql.keywords import (
    SYS_CLUSTER_TIME,
    SYS_CURRENT,
    SYS_DESCEND,
    SYS_KEEP,
    SYS_NOW,
    SYS_PRUNE,
    SYS_REMOVE,
    SYS_ROOT,
)
from easymql.meta import Grammar


class SystemVariables(Grammar):

    grammar = (
        SYS_CLUSTER_TIME
        | SYS_CURRENT
        | SYS_DESCEND
        | SYS_KEEP
        | SYS_NOW
        | SYS_PRUNE
        | SYS_REMOVE
        | SYS_ROOT
    )

    @classmethod
    def action(cls, token):
        return '$' + token[0]



from functools import reduce

from pyparsing import ParseException

from easymql.basics import *
from easymql.core import Forward
from easymql.datatypes.composite import Array
from easymql.datatypes.primary import Number, Boolean, String
from easymql.expressions.others import FieldPath
from easymql.keywords import *
from easymql.stages.parts import *
from easymql.utils import DelimitedList


class AddFields(Grammar):

    grammar = Suppress(ADD + FIELDS) + DelimitedList(Alias, min=1) + SEMICOLON

    @classmethod
    def action(cls, tokens):
        fields_dict = {}
        for t in tokens:
            fields_dict.update(t)
        return {"$addFields": fields_dict}


class BucketBy(Grammar):
    class Default(Grammar):
        grammar = DEFAULT + String

        @classmethod
        def action(cls, tokens):
            return {'default': tokens[-1]}

    class Output(Grammar):
        grammar = PROJECT + DelimitedList(ProjectAccumulator, min=1)

        @classmethod
        def action(cls, tokens):
            return {'output': reduce(lambda x, y: {**x, **y}, tokens[1:])}

    grammar = (
        BUCKET
        + BY
        + Expression
        + BOUNDARIES
        + Array
        + Optional(Default)
        + Optional(Output)
        + SEMICOLON
    )

    @classmethod
    def action(cls, tokens):
        doc = {'groupBy': tokens[2], 'boundaries': tokens[4]}
        for t in tokens[5:]:
            doc.update(t)
        return {'$bucket': doc}


class Count(Grammar):

    grammar = Suppress(COUNT + AS) + CollectionName + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$count": tokens[0]}


class GroupBy(Grammar):

    grammar = (
        GROUP
        + BY
        + Expression
        + Optional(PROJECT + DelimitedList(ProjectAccumulator, min=1))
        + SEMICOLON
    )

    @classmethod
    def action(cls, tokens):
        return {
            '$group': reduce(lambda x, y: {**x, **y}, [{'_id': tokens[2]}] + tokens[4:])
        }


class Limit(Grammar):

    grammar = Suppress(LIMIT) + Number + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$limit": tokens[0]}


class Lookup(Grammar):

    grammar = (
        LOOKUP
        + CollectionName
        + ON
        + Field
        + Literal('=')
        + Field
        + Optional(AS + Field)
        + SEMICOLON
    )

    @classmethod
    def action(cls, tokens):
        return {
            '$lookup': {
                'from': tokens[1],
                'localField': tokens[3],
                'foreignField': tokens[5],
                'as': tokens[-1] if len(tokens) == 8 else f'{tokens[1]}_docs',
            }
        }


class Match(Grammar):

    grammar = Suppress(MATCH) + Expression + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$match": {"$expr": tokens[0]}}


class OutputToDb(Grammar):

    grammar = Suppress(OUTPUT + TO) + DbCollectionPath + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$out": tokens[0]}


class Project(Grammar):
    class NewField(Grammar):
        grammar = Expression + AS + Field

        @classmethod
        def action(cls, tokens):
            return {tokens[-1]: tokens[0]}

    class ExcludeOrInclude(Grammar):
        grammar = Literal('+') + Field | Literal('-') + Field | Field

        @classmethod
        def action(cls, tokens):
            if len(tokens) == 1:
                return {tokens[0]: 1}
            else:
                return {tokens[-1]: 1 if tokens[0] == '+' else 0}

    # order matters
    element = NewField | ExcludeOrInclude
    grammar = Suppress(PROJECT) + DelimitedList(element, min=1) + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {'$project': reduce(lambda x, y: {**x, **y}, tokens)}


class Redact(Grammar):

    grammar = Suppress(REDACT) + Expression + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$redact": tokens[0]}


class ReplaceRoot(Grammar):
    grammar = Suppress(REPLACE + ROOT) + Expression + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$replaceRoot": {"newRoot": tokens[0]}}


class ReplaceWith(Grammar):
    grammar = Suppress(REPLACE + WITH) + Expression + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$replaceWith": tokens[0]}


class Sample(Grammar):

    grammar = Suppress(SAMPLE) + Number + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$sample": tokens[0]}


class Set(Grammar):

    grammar = Suppress(SET) + DelimitedList(Alias, min=1) + SEMICOLON

    @classmethod
    def action(cls, tokens):
        fields_dict = {}
        for t in tokens:
            fields_dict.update(t)
        return {"$set": fields_dict}


class Skip(Grammar):

    grammar = Suppress(SKIP | OFFSET) + Number + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$skip": tokens[0]}


class Sort(Grammar):
    class PairBySort(Grammar):
        grammar = Field + (ASC | DESC)

        @classmethod
        def action(cls, tokens):
            if tokens[1] == "ASC":
                tokens[1] = 1
            elif tokens[1] == 'DESC':
                tokens[1] = -1
            return {tokens[0]: tokens[1]}

    grammar = (
        Suppress(SORT | ORDER)
        + Suppress(BY)
        + DelimitedList(PairBySort, min=1)
        + SEMICOLON
    )

    @classmethod
    def action(cls, tokens):
        fields_dict = {}
        for t in tokens:
            fields_dict.update(t)
        return {"$sort": fields_dict}


class SortByCount(Grammar):

    grammar = Suppress(SORT + BY + COUNT) + Expression + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {"$sortByCount": tokens[0]}


class Unset(Grammar):

    grammar = Suppress(UNSET) + DelimitedList(Field, min=1) + SEMICOLON

    @classmethod
    def action(cls, tokens):
        token = []
        for i in tokens:
            token.append(i)
        return {"$unset": token}


class Unwind(Grammar):
    class ArrayIndex(Grammar):
        grammar = ARRAY + INDEX + AS + Field

        @classmethod
        def action(cls, tokens):
            return 'includeArrayIndex', tokens[-1]

    class PreserverNullEmptyArrays(Grammar):
        grammar = PRESERVE + NULL + EMPTY + ARRAYS + Boolean

        @classmethod
        def action(cls, tokens):
            return 'preserveNullAndEmptyArrays', tokens[-1]

    grammar = (
        UNWIND + FieldPath + (ArrayIndex | PreserverNullEmptyArrays)[0, 2] + SEMICOLON
    )

    @classmethod
    def action(cls, loc, tokens):
        # raise error if 2 options are given but both are the same
        if len(tokens) == 4 and tokens[2][0] == tokens[3][0]:
            if tokens[2][0] == 'includeArrayIndex':
                option = 'ARRAY INDEX AS'
            else:
                option = 'PRESERVE NULL EMPTY ARRAYS'
            raise ParseException('', loc=loc, msg=f"Duplicate option '{option}'")
        return {'$unwind': dict([('path', tokens[1])] + tokens[2:])}


class Merge(Grammar):
    class On(Grammar):
        grammar = ON + DelimitedList(Field, min=1)

        @classmethod
        def action(cls, tokens):
            print(tokens.asList())
            return {'on': tokens[1:]}

    class WhenMatched(Grammar):
        grammar = Suppress(WHEN + MATCHED + THEN) + (
            REPLACE
            | KEEP
            | MERGE
            | FAIL
            | LPAREN
            + (AddFields | Set | Project | Unset | ReplaceRoot | ReplaceWith)[1, ...]
            + RPAREN
        )

        @classmethod
        def action(cls, tokens):
            value = tokens.asList()
            if isinstance(tokens[0], str):
                if tokens[0] == 'KEEP':
                    value = 'keepExisting'
                else:
                    value = tokens[0].lower()
            return {'whenMatched': value}

    class WhenNotMatched(Grammar):
        grammar = Suppress(WHEN + NOT + MATCHED + THEN) + (INSERT | DISCARD | FAIL)

        @classmethod
        def action(cls, tokens):
            return {'whenNotMatched': tokens[0].lower()}

    grammar = (
        MERGE
        + INTO
        + DbCollectionPath
        + Optional(On)
        + Optional(WhenMatched)
        + Optional(WhenNotMatched)
        + SEMICOLON
    )

    @classmethod
    def action(cls, tokens):
        return {
            '$merge': reduce(
                lambda x, y: {**x, **y}, [{'into': tokens[2]}] + tokens[3:]
            )
        }


class Facet(Grammar):
    class NewPipeline(Grammar):
        """This should be updated with the new stages whenever get
        added that are supported in FACET"""

        allowed_stages = Forward()
        allowed_stages.set_name('pipeline')

        grammar = LPAREN + allowed_stages[1, ...] + RPAREN + Suppress(AS) + Field

        @classmethod
        def action(cls, tokens):
            print(tokens)
            return {tokens[-1]: tokens[0:-1]}

    grammar = FACET + DelimitedList(NewPipeline, min=1) + SEMICOLON

    @classmethod
    def action(cls, tokens):
        return {'$facet': reduce(lambda x, y: {**x, **y}, tokens[1:])}


class UnionWith(Grammar):
    """This should be updated with the new stages whenever get
    added that are supported in UNION WITH"""

    allowed_stages = Forward()
    allowed_stages.set_name('pipeline')

    grammar = (
        UNION
        + WITH
        + CollectionName
        + Optional(WITH + PIPELINE + LPAREN + allowed_stages[1, ...] + RPAREN)
        + SEMICOLON
    )

    allowed_stages <<= (
        AddFields
        | BucketBy
        | Count
        | Facet
        | GroupBy
        | Limit
        | Lookup
        | Match
        | Project
        | Redact
        | ReplaceRoot
        | ReplaceWith
        | Sample
        | Set
        | Skip
        | Sort
        | SortByCount
        | grammar
        | Unset
        | Unwind
    )

    @classmethod
    def action(cls, tokens):
        print(tokens)
        doc = {'coll': tokens[2]}
        if len(tokens) > 3:
            doc['pipeline'] = tokens[5:]
        return {'$unionWith': doc}


Facet.NewPipeline.allowed_stages <<= (
    AddFields
    | BucketBy
    | Count
    | GroupBy
    | Limit
    | Lookup
    | Match
    | Project
    | Redact
    | ReplaceRoot
    | ReplaceWith
    | Sample
    | Set
    | Skip
    | Sort
    | SortByCount
    | UnionWith
    | Unset
    | Unwind
)


Stages = (
    AddFields
    | BucketBy
    | Count
    | Facet
    | GroupBy
    | Limit
    | Lookup
    | Match
    | Merge
    | Project
    | OutputToDb
    | Redact
    | ReplaceRoot
    | ReplaceWith
    | Sample
    | Set
    | Skip
    | Sort
    | SortByCount
    | UnionWith
    | Unset
    | Unwind
)



from easymql.core import Literal

ABS = Literal("ABS")
ACOS = Literal('ACOS')
ACOSH = Literal('ACOSH')
ADD = Literal("ADD")
ADD_TO_SET = Literal("ADD_TO_SET")
ALL = Literal('ALL')
ANY = Literal('ANY')
ARRAY_ELEM_AT = Literal("ARRAY_ELEM_AT")
ARRAY_TO_OBJECT = Literal("ARRAY_TO_OBJECT")
ASIN = Literal('ASIN')
ASINH = Literal('ASINH')
ATAN = Literal('ATAN')
ATAN2 = Literal('ATAN2')
ATANH = Literal('ATANH')
AVG = Literal("AVG")
BINARY_SIZE = Literal("BINARY_SIZE")
BSON_SIZE = Literal("BSON_SIZE")
CEIL = Literal("CEIL")
CMP = Literal("CMP")
CONCAT = Literal('CONCAT')
CONCAT_ARRAYS = Literal("CONCAT_ARRAYS")
CONVERT = Literal('CONVERT')
COS = Literal('COS')
DATE = Literal('DATE')
DEGREES_TO_RADIANS = Literal('DEGREES_TO_RADIANS')
DIVIDE = Literal("DIVIDE")
EXP = Literal("EXP")
EXTRACT = Literal('EXTRACT')
FILTER = Literal("FILTER")
FIRST = Literal("FIRST")
FLOOR = Literal("FLOOR")
FORMAT_DATE = Literal('FORMAT_DATE')
IF = Literal("IF")
IF_NULL = Literal("IF_NULL")
IN = Literal("IN")
INDEX_OF_ARRAY = Literal("INDEX_OF_ARRAY")
INDEX_OF_BYTES = Literal('INDEX_OF_BYTES')
INDEX_OF_CP = Literal('INDEX_OF_CP')
ISO_WEEK_DATE = Literal('ISO_WEEK_DATE')
IS_ARRAY = Literal("IS_ARRAY")
IS_NUMBER = Literal('IS_NUMBER')
LAST = Literal("LAST")
LN = Literal("LN")
LOG = Literal("LOG")
LOG10 = Literal("LOG10")
LTRIM = Literal('LTRIM')
MAP = Literal("MAP")
MAX = Literal("MAX")
MERGE_OBJECTS = Literal("MERGE_OBJECTS")
MIN = Literal("MIN")
MOD = Literal("MOD")
MULTIPLY = Literal("MULTIPLY")
OBJECT_TO_ARRAY = Literal("OBJECT_TO_ARRAY")
PARSE_DATE = Literal('PARSE_DATE')
POW = Literal("POW")
PUSH = Literal("PUSH")
RADIANS_TO_DEGREES = Literal('RADIANS_TO_DEGREES')
RANDOM = Literal("RANDOM")
RANGE = Literal("RANGE")
REDUCE = Literal("REDUCE")
REGEX_FIND = Literal('REGEX_FIND')
REGEX_FIND_ALL = Literal('REGEX_FIND_ALL')
REGEX_MATCH = Literal('REGEX_MATCH')
REPLACE = Literal('REPLACE')
REPLACE_ALL = Literal('REPLACE_ALL')
REVERSE_ARRAY = Literal("REVERSE_ARRAY")
ROUND = Literal("ROUND")
RTRIM = Literal('RTRIM')
SAMPLE_RATE = Literal("SAMPLE_RATE")
SET_DIFFERENCE = Literal('SET_DIFFERENCE')
SET_EQUALS = Literal('SET_EQUALS')
SET_INTERSECTION = Literal('SET_INTERSECTION')
SET_IS_SUBSET = Literal('SET_IS_SUBSET')
SET_UNION = Literal('SET_UNION')
SIN = Literal('SIN')
SIZE = Literal("SIZE")
SLICE = Literal("SLICE")
SPLIT = Literal('SPLIT')
SQRT = Literal("SQRT")
STD_DEV_POP = Literal("STD_DEV_POP")
STD_DEV_SAMP = Literal("STD_DEV_SAMP")
STR_CASE_CMP = Literal('STR_CASE_CMP')
STR_LEN_BYTES = Literal('STR_LEN_BYTES')
STR_LEN_CP = Literal('STR_LEN_CP')
SUBSTR = Literal('SUBSTR')
SUBSTR_BYTES = Literal('SUBSTR_BYTES')
SUBSTR_CP = Literal('SUBSTR_CP')
SUBTRACT = Literal("SUBTRACT")
SUM = Literal("SUM")
TAN = Literal('TAN')
TO_BOOL = Literal('TO_BOOL')
TO_DATE = Literal('TO_DATE')
TO_DECIMAL = Literal('TO_DECIMAL')
TO_DOUBLE = Literal('TO_DOUBLE')
TO_INT = Literal('TO_INT')
TO_LONG = Literal('TO_LONG')
TO_LOWER = Literal('TO_LOWER')
TO_OBJECT_ID = Literal('TO_OBJECT_ID')
TO_STRING = Literal('TO_STRING')
TO_UPPER = Literal('TO_UPPER')
TRIM = Literal('TRIM')
TRUNC = Literal("TRUNC")
TYPE = Literal('TYPE')
ZIP = Literal("ZIP")



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.identifiers import *
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class Convert(Grammar):

    grammar = CONVERT + LPAREN + DelimitedList(expression_proxy, min=2, max=4) + RPAREN

    @classmethod
    def action(cls, tokens):
        return {'$convert': dict(zip(('input', 'to', 'onError', 'onNull'), tokens[1:]))}


class IsNumber(Grammar, ExpressionAction):

    grammar = IS_NUMBER + LPAREN + expression_proxy + RPAREN


class ToBool(Grammar, ExpressionAction):

    grammar = TO_BOOL + LPAREN + expression_proxy + RPAREN


class ToDate(Grammar, ExpressionAction):

    grammar = TO_DATE + LPAREN + expression_proxy + RPAREN


class ToDecimal(Grammar, ExpressionAction):

    grammar = TO_DECIMAL + LPAREN + expression_proxy + RPAREN


class ToDouble(Grammar, ExpressionAction):

    grammar = TO_DOUBLE + LPAREN + expression_proxy + RPAREN


class ToInt(Grammar, ExpressionAction):

    grammar = TO_INT + LPAREN + expression_proxy + RPAREN


class ToLong(Grammar, ExpressionAction):

    grammar = TO_LONG + LPAREN + expression_proxy + RPAREN


class ToObjectId(Grammar, ExpressionAction):

    grammar = TO_OBJECT_ID + LPAREN + expression_proxy + RPAREN


class ToString(Grammar, ExpressionAction):

    grammar = TO_STRING + LPAREN + expression_proxy + RPAREN


class Type(Grammar, ExpressionAction):

    grammar = TYPE + LPAREN + expression_proxy + RPAREN


TypeExpression = (
    Convert
    | IsNumber
    | ToBool
    | ToDate
    | ToDecimal
    | ToDouble
    | ToInt
    | ToLong
    | ToObjectId
    | ToString
    | Type
)



from easymql.basics import LPAREN, RPAREN, COMMA
from easymql.core import Keyword, MatchFirst, Optional, Suppress
from easymql.datatypes.primary import Null
from easymql.identifiers import DATE, ISO_WEEK_DATE, PARSE_DATE, FORMAT_DATE, EXTRACT
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class DateFunc(Grammar):

    grammar = (
        Suppress(DATE) + LPAREN + DelimitedList(expression_proxy, min=1, max=8) + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {
            '$dateFromParts': dict(
                zip(
                    [
                        'year',
                        'month',
                        'day',
                        'hour',
                        'minute',
                        'second',
                        'millisecond',
                        'timezone',
                    ],
                    tokens,
                )
            )
        }


class IsoWeekDateFunc(Grammar):

    grammar = (
        Suppress(ISO_WEEK_DATE)
        + LPAREN
        + DelimitedList(expression_proxy, min=1, max=8)
        + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {
            '$dateFromParts': dict(
                zip(
                    [
                        'isoWeekYear',
                        'isoWeek',
                        'isoDayOfWeek',
                        'hour',
                        'minute',
                        'second',
                        'millisecond',
                        'timezone',
                    ],
                    tokens,
                )
            )
        }


class ParseDate(Grammar):

    grammar = (
        PARSE_DATE + LPAREN + DelimitedList(expression_proxy, min=1, max=3) + RPAREN
    )

    @staticmethod
    def action(tokens):
        print(tokens)
        doc = {'dateString': tokens[1]}
        # Add 'format' to result only if it is not 'null'
        if len(tokens) >= 3 and not isinstance(tokens[2], Null):
            doc['format'] = tokens[2]
        if len(tokens) == 4:
            doc['timezone'] = tokens[3]
        return {'$dateFromString': doc}


class FormatDate(Grammar):

    grammar = (
        FORMAT_DATE + LPAREN + DelimitedList(expression_proxy, min=1, max=3) + RPAREN
    )

    @staticmethod
    def action(tokens):
        print(tokens)
        doc = {'date': tokens[1]}
        # Add 'format' to result only if it is not 'null'
        if len(tokens) >= 3 and not isinstance(tokens[2], Null):
            doc['format'] = tokens[2]
        if len(tokens) == 4:
            doc['timezone'] = tokens[3]
        return {'$dateToString': doc}


class Extract(Grammar):

    date_part_map = {
        'MILLISECOND': '$millisecond',
        'SECOND': '$second',
        'MINUTE': '$minute',
        'HOUR': '$hour',
        'DAY_OF_WEEK': '$dayOfWeek',
        'ISO_DAY_OF_WEEK': '$isoDayOfWeek',
        'DAY_OF_MONTH': '$day',
        'DAY_OF_YEAR': '$dayOfYear',
        'WEEK': '$week',
        'ISO_WEEK': '$isoWeek',
        'MONTH': '$month',
        'YEAR': '$year',
        'ISO_YEAR': '$isoYear',
    }
    grammar = (
        EXTRACT
        + LPAREN
        + MatchFirst(map(Keyword, date_part_map.keys()))
        + COMMA
        + expression_proxy
        + Optional(COMMA + expression_proxy)
        + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {
            Extract.date_part_map[tokens[1]]: dict(
                zip(('date', 'timezone'), tokens[2:])
            )
        }


DateExpression = DateFunc | IsoWeekDateFunc | ParseDate | FormatDate | Extract



from easymql.core import Suppress, Literal


(
    LPAREN,
    RPAREN,
    LBRACE,
    RBRACE,
    LBRACK,
    RBRACK,
    COLON,
    SEMICOLON,
    COMMA,
    PERIOD,
    HYPHEN,
    EQUALS,
) = map(Suppress, map(Literal, tuple('(){}[]:;,.-=')))



from easymql.actions import ExpressionAction, UnaryExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.datatypes.primary import Decimal
from easymql.identifiers import RANDOM, SAMPLE_RATE
from easymql.meta import Grammar


class Random(Grammar, ExpressionAction):

    grammar = RANDOM + LPAREN + RPAREN

    @staticmethod
    def action(tokens):
        return {'$rand': {}}


class SampleRate(Grammar, UnaryExpressionAction):

    grammar = SAMPLE_RATE + LPAREN + Decimal + RPAREN


MiscellaneousExpression = Random | SampleRate



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.identifiers import *
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class Abs(Grammar, ExpressionAction):

    grammar = ABS + LPAREN + expression_proxy + RPAREN


class Add(Grammar, ExpressionAction):

    grammar = ADD + LPAREN + DelimitedList(expression_proxy) + RPAREN


class Ceil(Grammar, ExpressionAction):

    grammar = CEIL + LPAREN + expression_proxy + RPAREN


class Divide(Grammar, ExpressionAction):

    grammar = DIVIDE + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN


class Exp(Grammar, ExpressionAction):

    grammar = EXP + LPAREN + expression_proxy + RPAREN


class Floor(Grammar, ExpressionAction):

    grammar = FLOOR + LPAREN + expression_proxy + RPAREN


class Ln(Grammar, ExpressionAction):

    grammar = LN + LPAREN + expression_proxy + RPAREN


class Log(Grammar, ExpressionAction):

    grammar = LOG + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN


class Log10(Grammar, ExpressionAction):

    grammar = LOG10 + LPAREN + expression_proxy + RPAREN


class Mod(Grammar, ExpressionAction):

    grammar = MOD + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN


class Multiply(Grammar, ExpressionAction):

    grammar = MULTIPLY + LPAREN + DelimitedList(expression_proxy) + RPAREN


class Pow(Grammar, ExpressionAction):

    grammar = POW + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN


class Round(Grammar, ExpressionAction):

    grammar = ROUND + LPAREN + DelimitedList(expression_proxy, min=1, max=2) + RPAREN


class Sqrt(Grammar, ExpressionAction):

    grammar = SQRT + LPAREN + expression_proxy + RPAREN


class Subtract(Grammar, ExpressionAction):

    grammar = SUBTRACT + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN


class Trunc(Grammar, ExpressionAction):

    grammar = TRUNC + LPAREN + DelimitedList(expression_proxy, min=1, max=2) + RPAREN


ArithmeticExpression = (
    Abs
    | Add
    | Ceil
    | Divide
    | Exp
    | Floor
    | Ln
    | Log
    | Log10
    | Mod
    | Multiply
    | Pow
    | Round
    | Sqrt
    | Subtract
    | Trunc
)



from easymql.actions import UnaryExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.expressions import Expression
from easymql.identifiers import *
from easymql.meta import Grammar


class AddToSet(Grammar, UnaryExpressionAction):

    grammar = ADD_TO_SET + LPAREN + Expression + RPAREN


class Avg(Grammar, UnaryExpressionAction):

    grammar = AVG + LPAREN + Expression + RPAREN


class First(Grammar, UnaryExpressionAction):

    grammar = FIRST + LPAREN + Expression + RPAREN


class Last(Grammar, UnaryExpressionAction):

    grammar = LAST + LPAREN + Expression + RPAREN


class Max(Grammar, UnaryExpressionAction):

    grammar = MAX + LPAREN + Expression + RPAREN


class MergeObjects(Grammar, UnaryExpressionAction):

    grammar = MERGE_OBJECTS + LPAREN + Expression + RPAREN


class Min(Grammar, UnaryExpressionAction):

    grammar = MIN + LPAREN + Expression + RPAREN


class Push(Grammar, UnaryExpressionAction):

    grammar = PUSH + LPAREN + Expression + RPAREN


class StdDevPop(Grammar, UnaryExpressionAction):

    grammar = STD_DEV_POP + LPAREN + Expression + RPAREN


class StdDevSamp(Grammar, UnaryExpressionAction):

    grammar = STD_DEV_SAMP + LPAREN + Expression + RPAREN


class Sum(Grammar, UnaryExpressionAction):

    grammar = SUM + LPAREN + Expression + RPAREN


class AccumulatorExpression(Grammar):

    name = 'accumulator_expression'
    grammar = (
        AddToSet
        | Avg
        | First
        | Last
        | Max
        | MergeObjects
        | Min
        | Push
        | StdDevPop
        | StdDevSamp
        | Sum
    )



from easymql.core import Keyword

ADD = Keyword("ADD")
ARRAY = Keyword('ARRAY')
ARRAYS = Keyword('ARRAYS')
AS = Keyword('AS')
ASC = Keyword('ASC')
BOUNDARIES = Keyword('BOUNDARIES')
BUCKET = Keyword('BUCKET')
BY = Keyword('BY')
CASE = Keyword("CASE")
COLL = Keyword('COLL')
COUNT = Keyword('COUNT')
DB = Keyword("DB")
DEFAULT = Keyword('DEFAULT')
DESC = Keyword('DESC')
DISCARD = Keyword('DISCARD')
ELSE = Keyword("ELSE")
EMPTY = Keyword('EMPTY')
END = Keyword("END")
FACET = Keyword('FACET')
FAIL = Keyword('FAIL')
FIELDS = Keyword('FIELDS')
GROUP = Keyword('GROUP')
INDEX = Keyword('INDEX')
INSERT = Keyword('INSERT')
INTO = Keyword('INTO')
KEEP = Keyword('KEEP')
LIMIT = Keyword('LIMIT')
LOOKUP = Keyword('LOOKUP')
MATCH = Keyword('MATCH')
MATCHED = Keyword('MATCHED')
MERGE = Keyword('MERGE')
NOT = Keyword('NOT')
NULL = Keyword('NULL')
OFFSET = Keyword('OFFSET')
ON = Keyword('ON')
ORDER = Keyword('ORDER')
OUTPUT = Keyword('OUTPUT')
PIPELINE = Keyword('PIPELINE')
PRESERVE = Keyword('PRESERVE')
PROJECT = Keyword('PROJECT')
REDACT = Keyword('REDACT')
REPLACE = Keyword('REPLACE')
ROOT = Keyword('ROOT')
SAMPLE = Keyword('SAMPLE')
SET = Keyword('SET')
SKIP = Keyword('SKIP')
SORT = Keyword('SORT')
THEN = Keyword("THEN")
TO = Keyword('TO')
UNION = Keyword('UNION')
UNSET = Keyword('UNSET')
UNWIND = Keyword('UNWIND')
WHEN = Keyword("WHEN")
WITH = Keyword('WITH')
false = Keyword('false')
null = Keyword('null')
true = Keyword('true')

# System variables
SYS_CLUSTER_TIME = Keyword('$CLUSTER_TIME')
SYS_CURRENT = Keyword('$CURRENT')
SYS_DESCEND = Keyword('$DESCEND')
SYS_KEEP = Keyword('$KEEP')
SYS_NOW = Keyword('$NOW')
SYS_PRUNE = Keyword('$PRUNE')
SYS_REMOVE = Keyword('$REMOVE')
SYS_ROOT = Keyword('$ROOT')



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.core import Suppress
from easymql.identifiers import *
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class ArrayElemAt(Grammar, ExpressionAction):

    grammar = (
        ARRAY_ELEM_AT + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN
    )


class ArrayToObject(Grammar, ExpressionAction):

    grammar = ARRAY_TO_OBJECT + LPAREN + expression_proxy + RPAREN


class ConcatArrays(Grammar, ExpressionAction):

    grammar = (
        CONCAT_ARRAYS
        + LPAREN
        + DelimitedList(expression_proxy, min=2, max=...)
        + RPAREN
    )


class Filter(Grammar, ExpressionAction):
    grammar = (
        Suppress(FILTER)
        + LPAREN
        + DelimitedList(expression_proxy, min=3, max=3)
        + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {"$filter": {"input": tokens[0], "as": tokens[1], "cond": tokens[2]}}


class First(Grammar, ExpressionAction):

    grammar = FIRST + LPAREN + expression_proxy + RPAREN


class In(Grammar, ExpressionAction):

    grammar = IN + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN


class IndexOfArray(Grammar, ExpressionAction):

    grammar = (
        INDEX_OF_ARRAY + LPAREN + DelimitedList(expression_proxy, min=2, max=4) + RPAREN
    )


class IsArray(Grammar, ExpressionAction):

    grammar = IS_ARRAY + LPAREN + expression_proxy + RPAREN


class Last(Grammar, ExpressionAction):

    grammar = LAST + LPAREN + expression_proxy + RPAREN


class Map(Grammar, ExpressionAction):
    grammar = (
        Suppress(MAP) + LPAREN + DelimitedList(expression_proxy, min=3, max=3) + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {"$map": {"input": tokens[0], "as": tokens[1], "in": tokens[2]}}


class ObjectToArray(Grammar, ExpressionAction):

    grammar = Suppress(OBJECT_TO_ARRAY) + LPAREN + expression_proxy + RPAREN

    @staticmethod
    def action(tokens):
        return {"$objectToArray": tokens[0]}


class Range(Grammar, ExpressionAction):

    grammar = RANGE + LPAREN + DelimitedList(expression_proxy, min=2, max=3) + RPAREN


class Reduce(Grammar, ExpressionAction):
    grammar = (
        Suppress(REDUCE)
        + LPAREN
        + DelimitedList(expression_proxy, min=3, max=3)
        + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {
            "$reduce": {"input": tokens[0], "initialValue": tokens[1], "in": tokens[2]}
        }


class ReverseArray(Grammar, ExpressionAction):

    grammar = REVERSE_ARRAY + LPAREN + expression_proxy + RPAREN


class Size(Grammar, ExpressionAction):

    grammar = SIZE + LPAREN + expression_proxy + RPAREN


class Slice(Grammar, ExpressionAction):

    grammar = SLICE + LPAREN + DelimitedList(expression_proxy, min=2, max=3) + RPAREN


class Zip(Grammar, ExpressionAction):
    grammar = (
        Suppress(ZIP) + LPAREN + DelimitedList(expression_proxy, min=1, max=3) + RPAREN
    )

    @staticmethod
    def action(tokens):

        result = {"inputs": tokens[0]}
        if len(tokens) >= 2:
            result["useLongestLength"] = tokens[1]

        if len(tokens) == 3:
            result["defaults"] = tokens[2]

        return {"$zip": result}


ArrayExpression = (
    ArrayElemAt
    | ArrayToObject
    | ConcatArrays
    | Filter
    | First
    | In
    | IndexOfArray
    | IsArray
    | Last
    | Map
    | ObjectToArray
    | Range
    | Reduce
    | ReverseArray
    | Size
    | Slice
    | Zip
)



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.identifiers import *
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class Sin(Grammar, ExpressionAction):

    grammar = SIN + LPAREN + expression_proxy + RPAREN


class Cos(Grammar, ExpressionAction):

    grammar = COS + LPAREN + expression_proxy + RPAREN


class Tan(Grammar, ExpressionAction):

    grammar = TAN + LPAREN + expression_proxy + RPAREN


class Asin(Grammar, ExpressionAction):

    grammar = ASIN + LPAREN + expression_proxy + RPAREN


class Acos(Grammar, ExpressionAction):

    grammar = ACOS + LPAREN + expression_proxy + RPAREN


class Atan(Grammar, ExpressionAction):

    grammar = ATAN + LPAREN + expression_proxy + RPAREN


class Atan2(Grammar, ExpressionAction):

    grammar = ATAN2 + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN


class Asinh(Grammar, ExpressionAction):

    grammar = ASINH + LPAREN + expression_proxy + RPAREN


class Acosh(Grammar, ExpressionAction):

    grammar = ACOSH + LPAREN + expression_proxy + RPAREN


class Atanh(Grammar, ExpressionAction):

    grammar = ATANH + LPAREN + expression_proxy + RPAREN


class DegreesToRadians(Grammar, ExpressionAction):

    grammar = DEGREES_TO_RADIANS + LPAREN + expression_proxy + RPAREN


class RadiansToDegrees(Grammar, ExpressionAction):

    grammar = RADIANS_TO_DEGREES + LPAREN + expression_proxy + RPAREN


TrigonometryExpression = (
    Sin
    | Cos
    | Tan
    | Asin
    | Acos
    | Atan
    | Atan2
    | Asinh
    | Acosh
    | Atanh
    | DegreesToRadians
    | RadiansToDegrees
)



from easymql.datatypes.composite import Array, Object
from easymql.datatypes.primary import Boolean, Date, Null, Number, String
from easymql.meta import Grammar


class DataType(Grammar):

    grammar = Array | Boolean | Date | Object | Number | String | Null



from pyparsing import ParseException

from easymql.exc import EasyMQLSyntaxError
from easymql.pipeline import Pipeline, encode


class EasyMQL:
    def parse(self, query_string):
        try:
            return encode(Pipeline.parse(query_string, explode=False))
        except ParseException as e:
            raise EasyMQLSyntaxError(query_string, str(e), e.lineno, e.col) from None



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.identifiers import *
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class Avg(Grammar, ExpressionAction):

    grammar = AVG + LPAREN + DelimitedList(expression_proxy, min=1, max=...) + RPAREN


class Max(Grammar, ExpressionAction):

    grammar = MAX + LPAREN + DelimitedList(expression_proxy, min=1, max=...) + RPAREN


class Min(Grammar, ExpressionAction):

    grammar = MIN + LPAREN + DelimitedList(expression_proxy, min=1, max=...) + RPAREN


class StdDevPop(Grammar, ExpressionAction):

    grammar = (
        STD_DEV_POP + LPAREN + DelimitedList(expression_proxy, min=1, max=...) + RPAREN
    )


class StdDevSamp(Grammar, ExpressionAction):

    grammar = (
        STD_DEV_SAMP + LPAREN + DelimitedList(expression_proxy, min=1, max=...) + RPAREN
    )


class Sum(Grammar, ExpressionAction):

    grammar = SUM + LPAREN + DelimitedList(expression_proxy, min=1, max=...) + RPAREN


AccumulatorExpression = Avg | Max | Min | StdDevPop | StdDevSamp | Sum



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.identifiers import BINARY_SIZE, BSON_SIZE
from easymql.meta import Grammar
from easymql.proxies import expression_proxy


class BinarySize(Grammar, ExpressionAction):

    grammar = BINARY_SIZE + LPAREN + expression_proxy + RPAREN


class BsonSize(Grammar, ExpressionAction):

    grammar = BSON_SIZE + LPAREN + expression_proxy + RPAREN


DataSizeExpression = BinarySize | BsonSize



from easymql.core import HashComment
from easymql.datatypes.primary import PrimaryDataType
from easymql.stages import Stages


def encode(item):
    """Convert Grammar object like Integer, String etc to python type"""
    if isinstance(item, list):
        for i, value in enumerate(item):
            item[i] = encode(value)
    elif isinstance(item, dict):
        for key, value in item.items():
            item[key] = encode(value)
    elif isinstance(item, PrimaryDataType):
        item = item.value
    return item


Pipeline = Stages[1, ...]
Pipeline.ignore(HashComment())



from pyparsing import (
    Empty as PpEmpty,
    Forward as PpForward,
    Keyword as PpKeyword,
    Literal as PpLiteral,
    Suppress as PpSuppress,
    Word as PpWord,
    QuotedString as PpQuotedString,
    Regex as PpRegex,
    Optional as PpOptional,
    White as PpWhite,
    oneOf,
    infixNotation as PpInfixNotation,
    opAssoc as OpAssoc,  # noqa
    MatchFirst as PpMatchFirst,
    And as PpAnd,
    pythonStyleComment,
    pyparsing_common,
)


class Adapter:
    def __init__(self, grammar):
        self.grammar = grammar
        try:
            self._set_parse_action(self.action)
        except AttributeError:
            pass
        try:
            self.set_name(str(self))
        except AttributeError:
            pass

    def _set_parse_action(self, action):
        try:
            self.grammar.setParseAction(action)
        except AttributeError:
            pass

    def set_name(self, name):
        self.grammar.setName(name)

    def __add__(self, other):
        return And([self, other])

    def __radd__(self, other):
        # return Adapter(self.grammar.__radd__(other._grammar))
        return other + self

    def __sub__(self, other):
        return Adapter(self.grammar.__sub__(other._grammar))

    def __rsub__(self, other):
        return other - self

    def __eq__(self, other):
        return self.grammar.__eq__(other._grammar)

    def __req__(self, other):
        return self == other

    def __ne__(self, other):
        return not (self == other)

    def __rne__(self, other):
        return not (self == other)

    def __getitem__(self, key):
        return MultipleMatch(self, key)

    def __mul__(self, other):
        return Adapter(self.grammar.__mul__(other._grammar))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __or__(self, other):
        return MatchFirst([self, other])

    def __ror__(self, other):
        return other | self

    # def __repr__(self):
    #     return f'{self.__class__.__name__}({self.value})'

    def __str__(self):
        return str(self.grammar)

    @property
    def _grammar(self):
        """Return PyParsing grammar contained in this instance."""
        return self.grammar

    def get_adapter_grammar(self):
        return self

    def parse(self, string, explode=True):
        result = self.grammar.parseString(string, parseAll=True).asList()
        if explode and len(result) == 1:
            return result.pop()
        else:
            return result

    def ignore(self, expr):
        return Adapter(self.grammar.ignore(expr._grammar))


class Keyword(Adapter):
    def __init__(self, match_string):
        self.match_string = match_string
        super(Keyword, self).__init__(PpKeyword(match_string))

    def __str__(self):
        return self.match_string


class Word(Adapter):
    def __init__(
        self,
        initChars,
        bodyChars=None,
        min=1,
        max=0,
        exact=0,
        asKeyword=False,
        excludeChars=None,
    ):
        grammar = PpWord(initChars, bodyChars, min, max, exact, asKeyword, excludeChars)
        super(Word, self).__init__(grammar)


class Suppress(Adapter):
    def __init__(self, expr):
        self.expr = expr
        super(Suppress, self).__init__(PpSuppress(expr._grammar))

    def __str__(self):
        return str(self.expr)


class QuotedString(Adapter):
    def __init__(
        self,
        quoteChar,
        escChar=None,
        escQuote=None,
        multiline=False,
        unquoteResults=True,
        endQuoteChar=None,
        convertWhitespaceEscapes=True,
    ):
        grammar = PpQuotedString(
            quoteChar,
            escChar,
            escQuote,
            multiline,
            unquoteResults,
            endQuoteChar,
            convertWhitespaceEscapes,
        )
        super(QuotedString, self).__init__(grammar)


class Forward(Adapter):
    def __init__(self, other=None):
        grammar = PpForward(other)
        super(Forward, self).__init__(grammar)

    def __lshift__(self, other):
        return Adapter(self.grammar.__lshift__(other._grammar))

    def __ilshift__(self, other):
        return Adapter(self.grammar.__ilshift__(other._grammar))


class Regex(Adapter):
    def __init__(self, pattern, flags=0, asGroupList=False, asMatch=False):
        grammar = PpRegex(pattern, flags, asGroupList, asMatch)
        super(Regex, self).__init__(grammar)


class Empty(Adapter):
    def __init__(self):
        super(Empty, self).__init__(PpEmpty())


class Literal(Adapter):
    def __init__(self, match_string):
        self.match_string = match_string
        super(Literal, self).__init__(PpLiteral(match_string))

    def __str__(self):
        return self.match_string


class Optional(Adapter):
    def __init__(self, expr):
        self.expr = expr
        super(Optional, self).__init__(PpOptional(expr._grammar))

    def __str__(self):
        return f'[ {self.expr._grammar} ]'


class White(Adapter):
    def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):
        super(White, self).__init__(PpWhite(ws, min, max, exact))


class OneOf(Adapter):
    def __init__(self, literals, case_less=False, use_regex=True, as_keyword=False):
        super(OneOf, self).__init__(oneOf(literals, case_less, use_regex, as_keyword))


class InfixExpression(Adapter):
    def __init__(
        self,
        base_expr,
        precedence_list,
        lparen=Suppress(Literal('(')),
        rparen=Suppress(Literal(')')),
    ):
        super(InfixExpression, self).__init__(
            PpInfixNotation(
                base_expr._grammar,
                [(p[0]._grammar, *p[1:]) for p in precedence_list],
                lparen._grammar,
                rparen._grammar,
            )
        )


class ParseExpression(Adapter):
    pass


class MatchFirst(ParseExpression):
    def __init__(self, exprs, savelist=False):
        self.exprs = exprs
        grammar = PpMatchFirst([expr._grammar for expr in exprs], savelist)
        super(MatchFirst, self).__init__(grammar)

    def _get_elements(self):
        """Flatten the nested MatchFirst objects and return as a list.

        { { A | B } | C } will become { A | B | C }.
        """
        res = []
        for expr in self.exprs:
            if isinstance(expr, MatchFirst):
                res += expr._get_elements()
            else:
                res.append(expr)
        return res

    def __str__(self):
        return '{ ' + ' | '.join(str(e._grammar) for e in self._get_elements()) + ' }'


class And(ParseExpression):
    def __init__(self, exprs, savelist=True):
        self.exprs = exprs
        grammar = PpAnd([expr._grammar for expr in exprs], savelist)
        super(And, self).__init__(grammar)

    def __str__(self):
        return ' '.join(str(e._grammar) for e in self.exprs)


class HashComment(Adapter):
    def __init__(self):
        super(HashComment, self).__init__(pythonStyleComment)


class MultipleMatch(Adapter):
    """Unlike others, this class does not use any specific PyParsing class.

    self.grammar here can be pyparsing.OneOrMore or pyparsing.ZeroOrMore.
    This class is created to override str behaviour of those classes.
    """

    def __init__(self, expr, key):
        self.expr = expr
        super(MultipleMatch, self).__init__(expr.grammar[key])

    def __str__(self):
        if isinstance(self.expr.get_adapter_grammar(), And):
            return f'{{ {self.expr._grammar} }}...'
        else:
            return f'{self.expr._grammar}...'


sci_real = Adapter(pyparsing_common.sci_real)
real = Adapter(pyparsing_common.real)
signed_integer = Adapter(pyparsing_common.signed_integer)



from easymql.core import QuotedString, Regex
from easymql.meta import Grammar


class FieldPath(Grammar):
    name = 'field_path'
    grammar = QuotedString(quoteChar="'", escChar='\\') | Regex(r'[\w.]+')

    @classmethod
    def action(cls, token):
        return '$' + token[0]



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.core import Optional, Suppress, White
from easymql.identifiers import IF, IF_NULL
from easymql.keywords import CASE, WHEN, END, ELSE, THEN
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class Cond(Grammar):

    grammar = (
        Suppress(IF) + LPAREN + DelimitedList(expression_proxy, min=3, max=3) + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {"$cond": {"if": tokens[0], "then": tokens[1], "else": tokens[2]}}


class IfNull(Grammar, ExpressionAction):

    grammar = IF_NULL + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN


class Case(Grammar):

    grammar = (
        Suppress(CASE)
        + DelimitedList(
            Suppress(WHEN) + expression_proxy + Suppress(THEN) + expression_proxy,
            delimiter=White(),
            min=1,
        )
        + Optional(Suppress(ELSE) + expression_proxy)
        + Suppress(END)
    )

    @staticmethod
    def action(tokens):
        res = {"$switch": {}}
        tokens = tokens.asList()

        if len(tokens) % 2 == 1:
            res["$switch"]["default"] = tokens.pop()

        i = 0
        branches = list()
        while i < len(tokens):
            b = {"case": tokens[i], "then": tokens[i + 1]}
            print(b)
            branches.append(b)
            i = i + 2
        res["$switch"]["branches"] = branches

        return res


ConditionalExpression = Cond | IfNull | Case



from easymql.core import Suppress, Regex, Optional, QuotedString
from easymql.expressions import Expression
from easymql.keywords import DB, COLL, AS
from easymql.meta import Grammar
from easymql.stages.groupacc import AccumulatorExpression


class CollectionName(Grammar):
    name = 'collection_name'
    grammar = QuotedString(quoteChar="'", escChar='\\') | Regex(r'[\w.]+')


class DbName(Grammar):
    name = 'db_name'
    grammar = QuotedString(quoteChar="'", escChar='\\') | Regex(r'\w+')


class DbCollectionPath(Grammar):
    name = 'collection_path'
    grammar = Optional(DB + DbName) + COLL + CollectionName

    @classmethod
    def action(cls, tokens):
        if len(tokens) == 2:
            return tokens[-1]
        else:
            return {'db': tokens[1], 'coll': tokens[-1]}


class Field(Grammar):
    name = 'field'
    grammar = QuotedString(quoteChar="'", escChar='\\') | Regex(r'[\w.]+')


class Alias(Grammar):

    grammar = Expression + Suppress(AS) + Field

    @classmethod
    def action(cls, tokens):
        return {tokens[1]: tokens[0]}


class ProjectAccumulator(Grammar):

    grammar = AccumulatorExpression + AS + Field

    @classmethod
    def action(cls, tokens):
        return {tokens[-1]: tokens[0]}



class EasyMQLError(Exception):
    pass


class DatePartOutOfRangeError(EasyMQLError):
    pass


class EasyMQLSyntaxError(EasyMQLError):
    def __init__(self, qstr, emsg, line_no, col, **kwargs):
        self.qstr = qstr
        self.emsg = emsg
        self.line_no = line_no
        self.col = col
        for k, v in kwargs:
            setattr(self, k, v)

    def __str__(self):
        query_string = '\n'.join(
            [
                line
                for no, line in enumerate(self.qstr.splitlines())
                if no < self.line_no
            ]
        )
        marker_padding = ' ' * (self.col - 1)
        return (
            f'\n'
            f'{query_string}\n'
            f'{marker_padding}^\n'
            f'Error at line: {self.line_no}, col: {self.col}\n'
            f'{self.emsg}'
        )



from easymql.core import InfixExpression, OpAssoc, OneOf, Keyword
from easymql.datatypes import DataType
from easymql.expressions.accumulators import AccumulatorExpression
from easymql.expressions.arithmetic import ArithmeticExpression
from easymql.expressions.array import ArrayExpression
from easymql.expressions.comparison import Cmp
from easymql.expressions.conditional import ConditionalExpression
from easymql.expressions.datasize import DataSizeExpression
from easymql.expressions.date import DateExpression
from easymql.expressions.miscellaneous import MiscellaneousExpression
from easymql.expressions.object import MergeObjects, ObjectToArray
from easymql.expressions.others import FieldPath
from easymql.expressions.set import SetExpression
from easymql.expressions.strings import StringExpression
from easymql.expressions.systemvariables import SystemVariables
from easymql.expressions.trigonometry import TrigonometryExpression
from easymql.expressions.type import TypeExpression
from easymql.meta import Grammar
from easymql.proxies import expression_proxy


class FuncExpression(Grammar):

    # Order matters. Especially FieldPath; it should be at the end.
    grammar = (
        DataType
        | AccumulatorExpression
        | ArithmeticExpression
        | ArrayExpression
        | ConditionalExpression
        | Cmp
        | DataSizeExpression
        | DateExpression
        | MiscellaneousExpression
        | MergeObjects
        | ObjectToArray
        | SetExpression
        | TrigonometryExpression
        | TypeExpression
        | StringExpression
        | SystemVariables
        | FieldPath
    )


def infix_action(tokens):
    tokens = tokens[0]

    # op : (op_name, -1 means any)
    # op_name = mongo operator name
    # nterms = no of expressions/terms the operator takes. -1 means any number of terms.
    op_map = {
        '+': ('$add', -1),
        '-': ('$subtract', 2),
        '*': ('$multiply', -1),
        '/': ('$divide', 2),
        '%': ('$mod', 2),
        '>': ('$gt', 2),
        '>=': ('$gte', 2),
        '<': ('$lt', 2),
        '<=': ('$lte', 2),
        '=': ('$eq', 2),
        '!=': ('$ne', 2),
        'AND': ('$and', -1),
        'OR': ('$or', -1),
    }

    res = tokens[0]
    previous_operator = None

    for i in range(1, len(tokens), 2):
        operator = tokens[i]
        right_arg = tokens[i + 1]
        op_name, nterms = op_map[operator]

        # to flatten operators with nterms = -1
        if nterms == -1 and operator == previous_operator:
            res[op_name].append(right_arg)
        else:  # nterms = 2
            res = {op_name: [res, right_arg]}

        previous_operator = operator

    return res


class Expression(Grammar):

    grammar = expression_proxy
    grammar <<= InfixExpression(
        FuncExpression,
        [
            (Keyword('NOT'), 1, OpAssoc.RIGHT, lambda token: {'$not': [token[0][-1]]}),
            (OneOf('* / %'), 2, OpAssoc.LEFT, infix_action),
            (OneOf('+ -'), 2, OpAssoc.LEFT, infix_action),
            (OneOf('< <= > >= = !='), 2, OpAssoc.LEFT, infix_action),
            (OneOf('AND OR', as_keyword=True), 2, OpAssoc.LEFT, infix_action),
        ],
    )



from easymql.core import (
    Suppress,
    Empty,
    Literal,
    Keyword,
    And,
    Adapter,
)


class DelimitedList(Adapter):
    def __init__(self, expr, delimiter=Literal(','), min=2, max=...):
        self.expr = expr
        super(DelimitedList, self).__init__(
            self._delimited_list(expr, delimiter, min, max)._grammar
        )

    def _delimited_list(self, expr, delimiter=Literal(','), min=2, max=...):
        if min is Ellipsis:
            raise ValueError('Min value should be definite. Use 0 instead.')
        if min == 0:
            if max == 0:
                return Empty()
            elif max == 1:
                return expr[0, 1]
            elif max is Ellipsis:
                return (expr + (Suppress(delimiter) + expr)[0, max])[0, 1]
            else:
                return (expr + (Suppress(delimiter) + expr)[0, max - 1])[0, 1]
        else:
            if max == 1:
                return expr
            elif max is Ellipsis:
                return expr + (Suppress(delimiter) + expr)[min - 1, max]
            else:
                return expr + (Suppress(delimiter) + expr)[min - 1, max - 1]

    def __str__(self):
        if isinstance(self.expr.get_adapter_grammar(), And):
            return f'{{ {self.expr._grammar} }}...'
        else:
            return f'{self.expr._grammar}...'


def safe_cast_int(int_str):
    try:
        return int(int_str)
    except Exception:
        return int_str



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.datatypes.composite import Array
from easymql.identifiers import *
from easymql.meta import Grammar
from easymql.utils import DelimitedList
from easymql.proxies import expression_proxy


class All(Grammar, ExpressionAction):

    grammar = ALL + LPAREN + expression_proxy + RPAREN

    @staticmethod
    def action(tokens):
        return {'$allElementsTrue': [tokens[-1]]}


class Any(Grammar, ExpressionAction):

    grammar = ANY + LPAREN + expression_proxy + RPAREN

    @staticmethod
    def action(tokens):
        return {'$anyElementTrue': [tokens[-1]]}


class SetDifference(Grammar, ExpressionAction):

    grammar = (
        SET_DIFFERENCE + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN
    )


class SetEquals(Grammar, ExpressionAction):

    grammar = SET_EQUALS + LPAREN + DelimitedList(expression_proxy) + RPAREN


class SetIntersection(Grammar, ExpressionAction):

    grammar = SET_INTERSECTION + LPAREN + DelimitedList(expression_proxy) + RPAREN


class SetIsSubset(Grammar, ExpressionAction):

    grammar = (
        SET_IS_SUBSET + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN
    )


class SetUnion(Grammar, ExpressionAction):

    grammar = SET_UNION + LPAREN + DelimitedList(expression_proxy) + RPAREN


SetExpression = (
    All | Any | SetDifference | SetEquals | SetIntersection | SetIsSubset | SetUnion
)



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.identifiers import MERGE_OBJECTS, OBJECT_TO_ARRAY
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class MergeObjects(Grammar, ExpressionAction):

    grammar = MERGE_OBJECTS + LPAREN + DelimitedList(expression_proxy) + RPAREN


class ObjectToArray(Grammar):

    grammar = OBJECT_TO_ARRAY + LPAREN + expression_proxy + RPAREN

    @classmethod
    def action(cls, tokens):
        return {'$objectToArray': tokens[1]}



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN
from easymql.identifiers import CMP
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class Cmp(Grammar, ExpressionAction):

    grammar = CMP + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN



from functools import reduce

from easymql.basics import *
from easymql.core import QuotedString
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class Array(Grammar):

    grammar = LBRACK + DelimitedList(expression_proxy, min=0, max=...) + RBRACK

    @classmethod
    def action(cls, tokens):
        return [tokens.asList()]


class KeyValuePair(Grammar):

    grammar = QuotedString(quoteChar='"', escChar='\\') + COLON + expression_proxy

    @classmethod
    def action(cls, tokens):
        return {tokens[0]: tokens[1]}


class Object(Grammar):

    grammar = LBRACE + DelimitedList(KeyValuePair, min=0, max=...) + RBRACE

    @classmethod
    def action(cls, tokens):
        return reduce(lambda e1, e2: {**e1, **e2}, tokens) if tokens else {}


Composite = Array | Object



from easymql.actions import ExpressionAction
from easymql.basics import LPAREN, RPAREN, COMMA
from easymql.core import Optional
from easymql.identifiers import *
from easymql.meta import Grammar
from easymql.proxies import expression_proxy
from easymql.utils import DelimitedList


class Concat(Grammar, ExpressionAction):

    grammar = CONCAT + LPAREN + DelimitedList(expression_proxy, min=2) + RPAREN


class IndexOfBytes(Grammar, ExpressionAction):

    grammar = (
        INDEX_OF_BYTES + LPAREN + DelimitedList(expression_proxy, min=2, max=4) + RPAREN
    )


class IndexOfCP(Grammar, ExpressionAction):

    grammar = (
        INDEX_OF_CP + LPAREN + DelimitedList(expression_proxy, min=2, max=4) + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {'$indexOfCP': tokens[1:]}


class Trim(Grammar, ExpressionAction):

    grammar = TRIM + LPAREN + DelimitedList(expression_proxy, min=1, max=2) + RPAREN

    @staticmethod
    def action(tokens):
        return {'$trim': dict(zip(('input', 'chars'), tokens[1:]))}


class LTrim(Grammar, ExpressionAction):

    grammar = LTRIM + LPAREN + DelimitedList(expression_proxy, min=1, max=2) + RPAREN

    @staticmethod
    def action(tokens):
        return {'$ltrim': dict(zip(('input', 'chars'), tokens[1:]))}


class RTrim(Grammar, ExpressionAction):

    grammar = RTRIM + LPAREN + DelimitedList(expression_proxy, min=1, max=2) + RPAREN

    @staticmethod
    def action(tokens):
        return {'$rtrim': dict(zip(('input', 'chars'), tokens[1:]))}


class RegexFind(Grammar):

    grammar = (
        REGEX_FIND
        + LPAREN
        + expression_proxy
        + COMMA
        + expression_proxy
        + Optional(COMMA + expression_proxy)
        + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {'$regexFind': dict(zip(('input', 'regex', 'options'), tokens[1:]))}


class RegexFindAll(Grammar):

    grammar = (
        REGEX_FIND_ALL
        + LPAREN
        + expression_proxy
        + COMMA
        + expression_proxy
        + Optional(COMMA + expression_proxy)
        + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {'$regexFindAll': dict(zip(('input', 'regex', 'options'), tokens[1:]))}


class RegexMatch(Grammar):

    grammar = (
        REGEX_MATCH
        + LPAREN
        + expression_proxy
        + COMMA
        + expression_proxy
        + Optional(COMMA + expression_proxy)
        + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {'$regexMatch': dict(zip(('input', 'regex', 'options'), tokens[1:]))}


class Replace(Grammar):

    grammar = REPLACE + LPAREN + DelimitedList(expression_proxy, min=3, max=3) + RPAREN

    @staticmethod
    def action(tokens):
        return {'$replaceOne': dict(zip(('input', 'find', 'replacement'), tokens[1:]))}


class ReplaceAll(Grammar):

    grammar = (
        REPLACE_ALL + LPAREN + DelimitedList(expression_proxy, min=3, max=3) + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {'$replaceAll': dict(zip(('input', 'find', 'replacement'), tokens[1:]))}


class Split(Grammar, ExpressionAction):

    grammar = SPLIT + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN


class StrLenBytes(Grammar, ExpressionAction):

    grammar = STR_LEN_BYTES + LPAREN + expression_proxy + RPAREN

    @staticmethod
    def action(tokens):
        return {'$strLenBytes': tokens[1]}


class StrLenCP(Grammar, ExpressionAction):

    grammar = STR_LEN_CP + LPAREN + expression_proxy + RPAREN

    @staticmethod
    def action(tokens):
        return {'$strLenCP': tokens[1]}


class StrCaseCmp(Grammar, ExpressionAction):

    grammar = (
        STR_CASE_CMP + LPAREN + DelimitedList(expression_proxy, min=2, max=2) + RPAREN
    )

    @classmethod
    def action(cls, tokens):
        return {'$strcasecmp': tokens[1:]}


class Substr(Grammar, ExpressionAction):

    grammar = SUBSTR + LPAREN + DelimitedList(expression_proxy, min=3, max=3) + RPAREN


class SubstrBytes(Grammar, ExpressionAction):

    grammar = (
        SUBSTR_BYTES + LPAREN + DelimitedList(expression_proxy, min=3, max=3) + RPAREN
    )


class SubstrCP(Grammar, ExpressionAction):

    grammar = (
        SUBSTR_CP + LPAREN + DelimitedList(expression_proxy, min=3, max=3) + RPAREN
    )

    @staticmethod
    def action(tokens):
        return {'$substrCP': tokens[1:]}


class ToLower(Grammar, ExpressionAction):

    grammar = TO_LOWER + LPAREN + expression_proxy + RPAREN


class ToUpper(Grammar, ExpressionAction):

    grammar = TO_UPPER + LPAREN + expression_proxy + RPAREN


StringExpression = (
    Concat
    | IndexOfBytes
    | IndexOfCP
    | Trim
    | LTrim
    | RTrim
    | RegexFind
    | RegexFindAll
    | RegexMatch
    | Replace
    | ReplaceAll
    | Split
    | StrCaseCmp
    | StrLenBytes
    | StrLenCP
    | Substr
    | SubstrBytes
    | SubstrCP
    | ToLower
    | ToUpper
)
